@@ -14,17 +14,16 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-import os
 import threading
 import time
 
 from concurrent import futures
-import fixtures
 import mock
 from six.moves.urllib import parse
 from testtools import matchers
 from testtools import testcase
 
+import tooz
 import tooz.coordination
 from tooz import tests
 
@@ -38,10 +37,7 @@ def try_to_lock_job(name, coord, url, member_id):
     return lock2.acquire(blocking=False)
 
 
-class TestAPI(tests.TestCaseSkipNotImplemented):
-
-    url = os.getenv("TOOZ_TEST_URL")
-
+class TestAPI(tests.TestWithCoordinator):
     def assertRaisesAny(self, exc_classes, callable_obj, *args, **kwargs):
         checkers = [matchers.MatchesException(exc_class)
                     for exc_class in exc_classes]
@@ -49,21 +45,6 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         callable_obj = testcase.Nullary(callable_obj, *args, **kwargs)
         self.assertThat(callable_obj, matcher)
 
-    def setUp(self):
-        super(TestAPI, self).setUp()
-        if self.url is None:
-            self.skipTest("No URL set for this driver")
-        self.useFixture(fixtures.NestedTempfile())
-        self.group_id = tests.get_random_uuid()
-        self.member_id = tests.get_random_uuid()
-        self._coord = tooz.coordination.get_coordinator(self.url,
-                                                        self.member_id)
-        self._coord.start()
-
-    def tearDown(self):
-        self._coord.stop()
-        super(TestAPI, self).tearDown()
-
     def test_connection_error_bad_host(self):
         if (tooz.coordination.Characteristics.DISTRIBUTED_ACROSS_HOSTS
            not in self._coord.CHARACTERISTICS):
@@ -78,13 +59,13 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_stop_first(self):
         c = tooz.coordination.get_coordinator(self.url,
                                               self.member_id)
-        self.assertRaises(tooz.coordination.ToozError,
+        self.assertRaises(tooz.ToozError,
                           c.stop)
 
     def test_create_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
 
     def test_get_lock_release_broken(self):
         name = tests.get_random_uuid()
@@ -124,10 +105,10 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_delete_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
         self._coord.delete_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertFalse(self.group_id in all_group_ids)
+        self.assertNotIn(self.group_id, all_group_ids)
         join_group = self._coord.join_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotCreated,
                           join_group.get)
@@ -150,13 +131,18 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
         member_list = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in member_list)
+        self.assertIn(self.member_id, member_list)
 
     def test_join_nonexistent_group(self):
         join_group = self._coord.join_group(self.group_id)
         self.assertRaises(tooz.coordination.GroupNotCreated,
                           join_group.get)
 
+    def test_join_group_create(self):
+        self._coord.join_group_create(self.group_id)
+        member_list = self._coord.get_members(self.group_id).get()
+        self.assertIn(self.member_id, member_list)
+
     def test_join_group_with_member_id_already_exists(self):
         self._coord.create_group(self.group_id).get()
         self._coord.join_group(self.group_id).get()
@@ -170,20 +156,20 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_leave_group(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
         self._coord.join_group(self.group_id).get()
         member_list = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in member_list)
+        self.assertIn(self.member_id, member_list)
         member_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in member_ids)
+        self.assertIn(self.member_id, member_ids)
         self._coord.leave_group(self.group_id).get()
         new_member_objects = self._coord.get_members(self.group_id).get()
         new_member_list = [member.member_id for member in new_member_objects]
-        self.assertTrue(self.member_id not in new_member_list)
+        self.assertNotIn(self.member_id, new_member_list)
 
     def test_leave_nonexistent_group(self):
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id not in all_group_ids)
+        self.assertNotIn(self.group_id, all_group_ids)
         leave_group = self._coord.leave_group(self.group_id)
         # Drivers raise one of those depending on their capability
         self.assertRaisesAny([tooz.coordination.MemberNotJoined,
@@ -193,7 +179,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
     def test_leave_group_not_joined_by_member(self):
         self._coord.create_group(self.group_id).get()
         all_group_ids = self._coord.get_groups().get()
-        self.assertTrue(self.group_id in all_group_ids)
+        self.assertIn(self.group_id, all_group_ids)
         leave_group = self._coord.leave_group(self.group_id)
         self.assertRaises(tooz.coordination.MemberNotJoined,
                           leave_group.get)
@@ -219,8 +205,7 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.join_group(group_id_test2).get()
         client2.join_group(group_id_test2).get()
         members_ids = self._coord.get_members(group_id_test2).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertEqual({self.member_id, member_id_test2}, members_ids)
 
     def test_get_member_capabilities(self):
         self._coord.create_group(self.group_id).get()
@@ -349,12 +334,12 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.join_group(self.group_id).get()
         client2.join_group(self.group_id).get()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertIn(member_id_test2, members_ids)
         client2.stop()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 not in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertNotIn(member_id_test2, members_ids)
 
     def test_timeout(self):
         if (tooz.coordination.Characteristics.NON_TIMEOUT_BASED
@@ -376,13 +361,17 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.join_group(self.group_id).get()
         client2.join_group(self.group_id).get()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertIn(member_id_test2, members_ids)
 
         # Watch the group, we want to be sure that when client2 is kicked out
         # we get an event.
         self._coord.watch_leave_group(self.group_id, self._set_event)
 
+        # Run watchers to be sure we initialize the member cache and we *know*
+        # client2 is a member now
+        self._coord.run_watchers()
+
         time.sleep(3)
         self._coord.heartbeat()
         time.sleep(3)
@@ -393,18 +382,20 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         while True:
             if self._coord.run_watchers():
                 break
-        self.assertTrue(self.member_id in members_ids)
-        self.assertTrue(member_id_test2 not in members_ids)
+        self.assertIn(self.member_id, members_ids)
+        self.assertNotIn(member_id_test2, members_ids)
         # Check that the event has been triggered
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberLeftGroup)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberLeftGroup)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
     def _set_event(self, event):
-        self.event = event
+        if not hasattr(self, "events"):
+            self.events = [event]
+        else:
+            self.events.append(event)
         return 42
 
     def test_watch_group_join(self):
@@ -420,26 +411,25 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         # Join the group
         client2.join_group(self.group_id).get()
         members_ids = self._coord.get_members(self.group_id).get()
-        self.assertTrue(member_id_test2 in members_ids)
+        self.assertIn(member_id_test2, members_ids)
         while True:
             if self._coord.run_watchers():
                 break
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberJoinedGroup)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberJoinedGroup)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
         # Stop watching
         self._coord.unwatch_join_group(self.group_id, self._set_event)
-        self.event = None
+        self.events = []
 
         # Leave and rejoin group
         client2.leave_group(self.group_id).get()
         client2.join_group(self.group_id).get()
         self._coord.run_watchers()
-        self.assertIsNone(self.event)
+        self.assertEqual([], self.events)
 
     def test_watch_leave_group(self):
         member_id_test2 = tests.get_random_uuid()
@@ -467,22 +457,21 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
             if self._coord.run_watchers():
                 break
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberLeftGroup)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberLeftGroup)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
         # Stop watching
         self._coord.unwatch_leave_group(self.group_id, self._set_event)
-        self.event = None
+        self.events = []
 
         # Rejoin and releave group
         client2.join_group(self.group_id).get()
         client2.leave_group(self.group_id).get()
         self._coord.run_watchers()
-        self.assertIsNone(self.event)
+        self.assertEqual([], self.events)
 
     def test_watch_join_group_disappear(self):
         if not hasattr(self._coord, '_destroy_group'):
@@ -502,9 +491,10 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         while True:
             if self._coord.run_watchers():
                 break
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberJoinedGroup)
-        self.event = None
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberJoinedGroup)
+        self.events = []
 
         # Force the group to disappear...
         self._coord._destroy_group(self.group_id)
@@ -513,8 +503,9 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
             if self._coord.run_watchers():
                 break
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.MemberLeftGroup)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.MemberLeftGroup)
 
     def test_watch_join_group_non_existent(self):
         self.assertRaises(tooz.coordination.GroupNotCreated,
@@ -546,7 +537,9 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         # Only works for clients that have access to the groups they are part
         # of, to ensure that after we got booted out by client3 that this
         # client now no longer believes its part of the group.
-        if hasattr(self._coord, '_joined_groups'):
+        if (hasattr(self._coord, '_joined_groups')
+           and (self._coord.run_watchers
+                == tooz.coordination.CoordinationDriverCachedRunWatchers.run_watchers)):  # noqa
             self.assertIn(self.group_id, self._coord._joined_groups)
             self._coord.run_watchers()
             self.assertNotIn(self.group_id, self._coord._joined_groups)
@@ -563,12 +556,11 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
         self._coord.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
     def test_run_for_election_multiple_clients(self):
         self._coord.create_group(self.group_id).get()
@@ -582,26 +574,26 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         client2.watch_elected_as_leader(self.group_id, self._set_event)
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
         self.assertEqual(self._coord.get_leader(self.group_id).get(),
                          self.member_id)
 
-        self.event = None
+        self.events = []
 
         self._coord.stop()
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
         self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+                         event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
         self.assertEqual(client2.get_leader(self.group_id).get(),
                          member_id_test2)
 
@@ -612,12 +604,12 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         self._coord.create_group(self.group_id).get()
 
         leader = self._coord.get_leader(self.group_id).get()
-        self.assertEqual(leader, None)
+        self.assertIsNone(leader)
 
         self._coord.join_group(self.group_id).get()
 
         leader = self._coord.get_leader(self.group_id).get()
-        self.assertEqual(leader, None)
+        self.assertIsNone(leader)
 
         # Let's get elected
         self._coord.watch_elected_as_leader(self.group_id, self._set_event)
@@ -638,36 +630,36 @@ class TestAPI(tests.TestCaseSkipNotImplemented):
         client2.watch_elected_as_leader(self.group_id, self._set_event)
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
 
-        self.event = None
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
+
+        self.events = []
 
         self._coord.stand_down_group_leader(self.group_id)
         client2.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(member_id_test2,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
 
-        self.event = None
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(member_id_test2, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
+
+        self.events = []
 
         client2.stand_down_group_leader(self.group_id)
         self._coord.run_watchers()
 
-        self.assertIsInstance(self.event,
-                              tooz.coordination.LeaderElected)
-        self.assertEqual(self.member_id,
-                         self.event.member_id)
-        self.assertEqual(self.group_id,
-                         self.event.group_id)
+        self.assertEqual(1, len(self.events))
+        event = self.events[0]
+
+        self.assertIsInstance(event, tooz.coordination.LeaderElected)
+        self.assertEqual(self.member_id, event.member_id)
+        self.assertEqual(self.group_id, event.group_id)
 
     def test_unwatch_elected_as_leader(self):
         # Create a group and add a elected_as_leader callback
